PROGRAM NO2AMF_AIRCRAFT_MAIN

      USE VLIDORT_PARS

      USE VLIDORT_BRDFSup_def
      USE VLIDORT_LSBRDFSup_def

      USE VLIDORT_Inputs_def
      USE VLIDORT_LinInputs_def
      USE VLIDORT_Outputs_def
      USE VLIDORT_LCOutputs_def
      USE VLIDORT_LPOutputs_def
      USE VLIDORT_LSOutputs_def
      USE VLIDORT_LinOutputs_def

      USE VLIDORT_AUX
      USE VLIDORT_INPUTS
      USE VLIDORT_MASTERS
      USE VLIDORT_L_INPUTS
      USE VLIDORT_L_MASTERS

      IMPLICIT NONE

!  VLIDORT input structures

      TYPE(Fixed_Inputs_Boolean_def)     :: VLIDORT_FBoolean_inputs
      TYPE(Fixed_Inputs_Control_def)     :: VLIDORT_FControl_inputs
      TYPE(Modified_Inputs_Boolean_def)  :: VLIDORT_MBoolean_inputs
      TYPE(Modified_Inputs_Control_def)  :: VLIDORT_MControl_inputs

      TYPE(Fixed_Inputs_Sunrays_def)     :: VLIDORT_Beam_inputs
      TYPE(Fixed_Inputs_UserVAL_def)     :: VLIDORT_User_inputs
      TYPE(Fixed_Inputs_Chapman_def)     :: VLIDORT_Chapman_inputs

      TYPE(Fixed_Inputs_Optical_def)     :: VLIDORT_Optical_inputs

      TYPE(Fixed_Inputs_Write_def)       :: VLIDORT_Write_inputs

!  VLIDORT linearized input structures

      TYPE(Fixed_Inputs_LinControl_def)    :: VLIDORT_FLinControl_inputs
      TYPE(Modified_Inputs_LinControl_def) :: VLIDORT_MLinControl_inputs
      TYPE(Inputs_LinIOPs_Atmos_def)       :: VLIDORT_LinOptical_inputs

!  VLIDORT BRDF input structures

      TYPE(BRDF_Surface_def)             :: VLIDORT_BRDF_inputs
      TYPE(LSBRDF_Surface_def)           :: VLIDORT_LSBRDF_inputs

!  VLIDORT output structures

      TYPE(Input_Exception_Handling_def) :: VLIDORT_InputStatus
      TYPE(Outputs_Main_def)             :: VLIDORT_Outputs
      TYPE(Exception_Handling_def)       :: VLIDORT_Status

!  VLIDORT linearized output structures

      TYPE(LCOutputs_Main_def)           :: VLIDORT_LCOutputs
      TYPE(LPOutputs_Main_def)           :: VLIDORT_LPOutputs
      TYPE(LinOutputs_Main_def)          :: VLIDORT_LinOutputs
      TYPE(LSOutputs_Main_def)           :: VLIDORT_LSOutputs

!  Flag for opening error output file

      LOGICAL         :: OPENFILEFLAG

!  Original LIDORT 2.3 variables (Old tool)
!  ========================================

!       warning: max number of layers < ( MAXLAYER = 120 )

      double precision sza,cloud_ssa
      parameter (cloud_ssa=0.9999)
      character csza*2
!      parameter (sza=60.0D+00,csza='60')
      parameter (sza=46.0D+00,csza='46')
      double precision salbedo,cld_toa	! simulated TOA reflectance at surface albedo (salbedo) 
      parameter (salbedo=0.05)
      character csp*4		! this defines vertical optical input file for given species 9so2,ozn, no2) and wavelength
      parameter (csp='no2.')

      double precision isave2                                      

      character*4 stcloud
      character*7 fout    

      integer          i, j, ui, n, ndum, idum, l, nmom, k, nf, na, nlayers, nmoments_input
      integer ip


!  Masking arrays

      INTEGER ::          NGREEKMAT_ENTRIES, NSTOKES
      INTEGER ::          CK, GK, CMASK(8), GMASK(8), SMASK(8)

!      GMASK = (/  1, 2, 5, 6, 11, 12, 15, 16 /)
!      CMASK = (/  1, 5, 5, 2, 3, 6, 6, 4 /)     !  OUTPUT OF RTS MIE NEW CODE
!      SMASK = (/  1, -1, -1, 1, 1,  1, -1, 1 /) !  NEW MIE CODE

!      CMASK = (/  1, 2, 2, 3, 4, 5, 5, 6 /)     !  OUTPUT OF RTS MIE. OLD CODE
!      SMASK = (/  1, -1, -1, 1, 1, -1, 1, 1 /)  !  OLD MIE CODE

!  Atmosphere stuff

      double precision wvlth,psrfc,xno2,tautot,sk,mom,beta2


      double precision tray_tot,tno2_tot,tcl_tot,taer_tot

      double precision dum, tot_tau_data(100), gas_tau_data(100), aer_tau(100), cld_tau(100), ray_tau_data(100)
      double precision depol, ray_2mom, tau_scat, amom(6,500)
      double precision gas_tau, tot_tau, ray_tau, tottau_save(100), omega
      double precision pres_data(100),height_grid(0:100), pres_save(100),tauscat_save(100)
      double precision epsilon, epsfac

!  Code for setting up the Aircraft situation

      double precision p_aircraft, h_aircraft, ratio, gradient
      integer          nac
!  flag for skipping Rayleigh TSb calculations

      logical, parameter :: skip_TSb_calculation = .true.
      double precision PROBLEM_RAY(6,0:2)
!  START OF CODE
!  =============

!  1. INITIAL SECTION
!  ==================

!  Get Aircraft Pressure

      p_aircraft = 110.0d0

!  Initialize error handling Open-file flag

      OPENFILEFLAG = .false.

!  Read the VLIDORT control Configuration file

      CALL VLIDORT_input_master ( 'NO2AMF_ReadInput_VLIDORT.cfg', &
        VLIDORT_FBoolean_inputs,  &  ! OUTPUTS
        VLIDORT_MBoolean_inputs,  &  ! OUTPUTS
        VLIDORT_FControl_inputs,  &  ! OUTPUTS
        VLIDORT_MControl_inputs,  &  ! OUTPUTS
        VLIDORT_Beam_inputs,      &  ! OUTPUTS
        VLIDORT_User_inputs,      &  ! OUTPUTS
        VLIDORT_Chapman_inputs,   &  ! OUTPUTS
        VLIDORT_Optical_inputs,   &  ! OUTPUTS
        VLIDORT_Write_inputs,     &  ! OUTPUTS
        VLIDORT_InputStatus )        ! OUTPUTS
!  Exception handling on this file-read operation. Abort if error.

      IF ( VLIDORT_InputStatus%TS_STATUS_INPUTREAD.ne.vlidort_success ) then
        open(1,file = 'NO2_ReadInput.log', status = 'unknown')
        WRITE(1,*)' FATAL:   Wrong input from VLIDORT input file-read'
        WRITE(1,*)'  ------ Here are the messages and actions '
        write(1,'(A,I3)')'    ** Number of messages = ',VLIDORT_InputStatus%TS_NINPUTMESSAGES
        DO N = 1, VLIDORT_InputStatus%TS_NINPUTMESSAGES
         NF = LEN_STRING(VLIDORT_InputStatus%TS_INPUTMESSAGES(N))
         NA = LEN_STRING(VLIDORT_InputStatus%TS_INPUTACTIONS(N))
         write(1,'(A,I3,A,A)')'Message # ',N,' : ',VLIDORT_InputStatus%TS_INPUTMESSAGES(N)(1:NF)
         write(1,'(A,I3,A,A)')'Action  # ',N,' : ',VLIDORT_InputStatus%TS_INPUTACTIONS(N)(1:NA)
        ENDDO
        close(1)
        STOP'Read-input fail: Look at file NO2_ReadInput.log'
      ENDIF

!  Zero some input variables (not required in this application)

      VLIDORT_FControl_inputs%TS_NLAYERS_NOMS    = ZERO
      VLIDORT_FControl_inputs%TS_NLAYERS_CUTOFF  = ZERO

      VLIDORT_Chapman_inputs%TS_PRESSURE_GRID    = ZERO
      VLIDORT_Chapman_inputs%TS_TEMPERATURE_GRID = ZERO
      VLIDORT_Chapman_inputs%TS_FINEGRID         = ZERO

      VLIDORT_Optical_inputs%TS_THERMAL_BB_INPUT = ZERO
      VLIDORT_Optical_inputs%TS_SURFACE_BB_INPUT = ZERO

      VLIDORT_Optical_inputs%TS_LTE_DELTAU_VERT_INPUT = ZERO
      VLIDORT_Optical_inputs%TS_LTE_THERMAL_BB_INPUT  = ZERO

      VLIDORT_BRDF_inputs%TS_EXACTDB_BRDFUNC = ZERO
      VLIDORT_BRDF_inputs%TS_BRDF_F_0        = ZERO
      VLIDORT_BRDF_inputs%TS_BRDF_F          = ZERO
      VLIDORT_BRDF_inputs%TS_USER_BRDF_F_0   = ZERO
      VLIDORT_BRDF_inputs%TS_USER_BRDF_F     = ZERO
      VLIDORT_BRDF_inputs%TS_EMISSIVITY      = ZERO
      VLIDORT_BRDF_inputs%TS_USER_EMISSIVITY = ZERO

      VLIDORT_LSBRDF_inputs%TS_LS_EXACTDB_BRDFUNC = ZERO
      VLIDORT_LSBRDF_inputs%TS_LS_BRDF_F_0        = ZERO
      VLIDORT_LSBRDF_inputs%TS_LS_BRDF_F          = ZERO
      VLIDORT_LSBRDF_inputs%TS_LS_USER_BRDF_F_0   = ZERO
      VLIDORT_LSBRDF_inputs%TS_LS_USER_BRDF_F     = ZERO
      VLIDORT_LSBRDF_inputs%TS_LS_EMISSIVITY      = ZERO
      VLIDORT_LSBRDF_inputs%TS_LS_USER_EMISSIVITY = ZERO

!  Open Data file with atmospheric data 

      open(1,file='Layers_'//csp//'tc05',status='old')
      read(1,*)
      read(1,*) wvlth,psrfc,depol,xno2,tautot,dum,nlayers
      read(1,*)
      read(1,*)tray_tot,taer_tot,tno2_tot,tcl_tot,dum,dum,idum
      read(1,*)
      read(1,*)dum,dum
      read(1,*)
      do n = 1, nlayers
        read(1,*)idum,HEIGHT_GRID(n),pres_data(n), ray_tau_data(n), aer_tau(n), &
                   gas_tau_data(n), cld_tau(n), tot_tau_data(n)
!        print*, HEIGHT_GRID(n),pres_data(n), ray_tau_data(n), aer_tau(n), &
!                   gas_tau_data(n), cld_tau(n), tot_tau_data(n)
      enddo
      close(1)


!  Set TOA. 
!     ######## WARNING - this choice has a small effect on the results

      height_grid(0) = 50.0d0

!  Open cloud-moments file

      open(1,file='C1_pmom_lesha',status='old')
      read(1,*)nmom
      do n = 0, nmom - 1
        read(1,*) ndum, amom(1,n+1)
      enddo
      close(1)

!  Set masking limits

      nstokes = VLIDORT_FControl_inputs%TS_nstokes
      if ( nstokes .eq. 1 ) ngreekmat_entries = 1
      if ( nstokes .eq. 3 ) ngreekmat_entries = 5
      if ( nstokes .eq. 4 ) ngreekmat_entries = 8

!  Set VLIDORT Flux-factor and solar angle

      VLIDORT_Beam_inputs%TS_FLUX_FACTOR               = PIE
      VLIDORT_Beam_inputs%TS_SZANGLES(1)              = sza

!  Set VLIDORT Nlayers and height grid

      VLIDORT_FControl_inputs%TS_nlayers               = nlayers
      VLIDORT_Chapman_inputs%TS_height_grid(0:nlayers) = height_grid(0:nlayers)
      VLIDORT_User_inputs%TS_geometry_specheight       = height_grid(nlayers)
!  Rayleigh scattering law

      ray_2mom = ( one - depol ) / ( two + depol )
      beta2 = ray_2mom
      PROBLEM_RAY = 0.0d0
      PROBLEM_RAY(1,0) =  1.D0
      PROBLEM_RAY(3,1) =  3.D0 * ( 1.D0 - 2.D0*DEPOL ) / (2.D0 + DEPOL )
      PROBLEM_RAY(1,2) =  beta2
      PROBLEM_RAY(5,2) =  - DSQRT(6.D0) * beta2
      PROBLEM_RAY(2,2) =  6.D0 * beta2

!  Additional task. Find the output level (partial layer output at P_aircraft)
!  ----------------

!    Trawl through pressures to find the layer containing aircraft

      n = 1
      do while (p_aircraft.ge.pres_data(n) )
         n = n + 1
      enddo
      Nac = n
      write(*,*)nac
      gradient   = (height_grid(nac)-height_grid(nac-1))/ dlog(pres_data(nac)/pres_data(nac-1))
      H_aircraft = height_grid(nac-1) + gradient * dlog(p_aircraft/pres_data(nac-1))

!  set the VLIDORT User level
!  set the VLIDORT User level

      ratio = (H_aircraft - height_grid(nac-1) ) /  (height_grid(nac)-height_grid(nac-1))
      VLIDORT_User_inputs%TS_USER_LEVELS(1) = DBLE(Nac-1) + ratio
      write(*,*)'Aircraft position: ',nac,h_aircraft,VLIDORT_User_inputs%TS_USER_LEVELS(1) 
!      pause'Aircraft setup'



!  3. Cloud WF calculation for Ps(nprs)                   
!     =================================

!  Start pressure loop

      do ip = 1, 1
!      do ip = 1, nprs

!  3A. Baseline VLIDORT calculation with Jacobians
!  ----------------------------------------------

!  set VLIDORT Deltam scaling and SSCORR_Nadir, Set Nmoments = 324

        VLIDORT_MControl_inputs%TS_ngreek_moments_input     = 2
!        VLIDORT_MControl_inputs%TS_ngreek_moments_input     = 324
        VLIDORT_MBoolean_inputs%TS_DO_DELTAM_SCALING  = .false.
        VLIDORT_MBoolean_inputs%TS_DO_SSCORR_NADIR    = .true.
        VLIDORT_MBoolean_inputs%TS_DO_RAYLEIGH_ONLY   = .true.

!  Set profile linearization flag

        VLIDORT_MLinControl_inputs%TS_DO_PROFILE_LINEARIZATION = .true.
        VLIDORT_MLinControl_inputs%TS_DO_COLUMN_LINEARIZATION  = .false.
        VLIDORT_MLinControl_inputs%TS_DO_SURFACE_LINEARIZATION = .false.
        VLIDORT_MLinControl_inputs%TS_DO_ATMOS_LINEARIZATION  = .true.

!  set profile linearization control

        VLIDORT_FLinControl_inputs%TS_LAYER_VARY_FLAG          = .true.
        VLIDORT_FLinControl_inputs%TS_LAYER_VARY_NUMBER        = 1

!  Zero unwanted outputs
!  Zero unwanted outputs

        VLIDORT_FLinControl_inputs%TS_N_TOTALCOLUMN_WFS        = 0
        VLIDORT_FLinControl_inputs%TS_N_SURFACE_WFS            = 0
        VLIDORT_LinOptical_inputs%TS_L_greekmat_total_input = zero

!  Set the albedo

        VLIDORT_Optical_inputs%TS_LAMBERTIAN_ALBEDO = salbedo

!  write header

        WRITE(57,*) 'wvlth psrfc  rho  xNO2[atmcm] tautot deltau nolyr'
	WRITE(57,*) wvlth,psrfc,depol,xno2,tautot,dum,nlayers
        WRITE(57,*) 'tray   taer   tno2   tcld  '
        WRITE(57,*)tray_tot,taer_tot,tno2_tot,tcl_tot
	WRITE(57,* )         'Surface albedo         TOA radiance' 

!  Set atmospheric optical properties and their linearizations

        do n = 1, nlayers
  	  pres_save(n) = pres_data(n)
	  tot_tau = tot_tau_data(n)
          ray_tau = ray_tau_data(n)			! Scale Rayleigh tau according to surface pressure
          tau_scat = ray_tau + aer_tau(n) + cld_tau(n)
          omega    = tau_scat / tot_tau
          tottau_save(n)   = tot_tau
          tauscat_save(n)  = tau_scat
          if ( omega .ge. one ) omega = 0.99999d0
          VLIDORT_Optical_inputs%TS_deltau_vert_input(n) = tot_tau
          VLIDORT_Optical_inputs%TS_omega_total_input(n) = omega

          DO K = 1, ngreekmat_entries
             GK = gmask(k);  sk = dble(smask(k)) ;  ck = cmask(k)
             DO L = 0, 2
                MOM = ( ray_tau*PROBLEM_RAY(ck,L) + CLD_TAU(N)*SK*AMOM(CK,L+1) ) / tau_scat
                VLIDORT_Optical_inputs%TS_GREEKMAT_TOTAL_INPUT(L,N,GK) = MOM
             ENDDO
             DO L = 3, VLIDORT_MControl_inputs%TS_ngreek_moments_input
                MOM = cld_tau(n) * sk * amom(ck,l+1) / tau_scat
                VLIDORT_Optical_inputs%TS_GREEKMAT_TOTAL_INPUT(L,N,GK) = MOM
             ENDDO
          ENDDO

          VLIDORT_Optical_inputs%TS_GREEKMAT_TOTAL_INPUT(0,n,1) = 1.0d0

          VLIDORT_LinOptical_inputs%TS_L_deltau_vert_input(1,n) = + gas_tau_data(n) / tot_tau
          VLIDORT_LinOptical_inputs%TS_L_omega_total_input(1,n) = - gas_tau_data(n) / tot_tau

        enddo

!  VLIDORT call ( Linearized Master for Profile (P) and Surface (S) Jacobians )
!  VLIDORT call ( Linearized Master for Profile (P) and Surface (S) Jacobians )

        CALL VLIDORT_L_master ( &
          VLIDORT_FBoolean_inputs,   &  ! INPUTS
          VLIDORT_MBoolean_inputs,   &  ! INPUTS (possibly modified)
          VLIDORT_FControl_inputs,   &  ! INPUTS
          VLIDORT_Mcontrol_inputs,   &  ! INPUTS (possibly modified)
          VLIDORT_Beam_inputs,       &  ! INPUTS
          VLIDORT_User_inputs,       &  ! INPUTS
          VLIDORT_Chapman_inputs,    &  ! INPUTS
          VLIDORT_Optical_inputs,    &  ! INPUTS
          VLIDORT_Write_inputs,       &
          VLIDORT_FLinControl_inputs, &
          VLIDORT_MLinControl_inputs, &
          VLIDORT_LinOptical_inputs,  &
          VLIDORT_BRDF_inputs,        &
          VLIDORT_LSBRDF_inputs,      &
          VLIDORT_Outputs,            &
          VLIDORT_LCOutputs,          &
          VLIDORT_LPOutputs,          &
          VLIDORT_LSOutputs,          &
          VLIDORT_LinOutputs,         &
          VLIDORT_Status )              ! Outputs

!  save baseline radiances

        isave2 = VLIDORT_Outputs%TS_STOKES(1,1,1,UPIDX)
     write(*,*)'isave2',isave2,VLIDORT_LPOutputs%TS_PROFILEWF(1,1,1,1,1,1)

!  Exception handling

        OPENFILEFLAG = .false.
        CALL VLIDORT_WRITE_STATUS ( &
          'NO2_AMF_Execution.log', 35, OPENFILEFLAG, &
          VLIDORT_Status%TS_STATUS_INPUTCHECK,   &
          VLIDORT_Status%TS_NCHECKMESSAGES,      &
          VLIDORT_Status%TS_CHECKMESSAGES,       &
          VLIDORT_Status%TS_ACTIONS,             &
          VLIDORT_Status%TS_STATUS_CALCULATION,  &
          VLIDORT_Status%TS_MESSAGE,             &
          VLIDORT_Status%TS_TRACE_1,             &
          VLIDORT_Status%TS_TRACE_2,             &
          VLIDORT_Status%TS_TRACE_3 ) 

!  Stop if calculation error

        IF ( VLIDORT_Status%TS_STATUS_CALCULATION.ne.vlidort_success ) then
          write(*,*)'Baseline Linearization Run aborted for pressure cycel number ',ip,'; Look at error Log file'
          Stop
        ENDIF

!    pause 'Done baseline'
!    pause 'Done baseline'

!  3B. Finite difference weighting functions
!  -----------------------------------------

!  Turn off linearization control

        VLIDORT_MLinControl_inputs%TS_DO_ATMOS_LINEARIZATION   = .false.
        VLIDORT_MLinControl_inputs%TS_DO_PROFILE_LINEARIZATION = .false.
        VLIDORT_FLinControl_inputs%TS_LAYER_VARY_FLAG          = .false.
        VLIDORT_FLinControl_inputs%TS_LAYER_VARY_NUMBER        = 0

!  Set Finite-difference perturbation

        epsilon = 0.001d0
        epsfac  = 1.0d0 + epsilon

!  Loop over atmospheric layers 

        do k = 1, nlayers

          write(*,*)'doing perturbation ',k

!  reset atmospheric properties according to perturbed gas amount

          do n = 1, nlayers
            if ( k.eq.n ) then
              gas_tau = gas_tau_data(n) * epsfac
              tot_tau = tottau_save(n) + gas_tau_data(n) * epsilon
            else
              gas_tau = gas_tau_data(n)
              tot_tau = tottau_save(n)
            endif
            omega = tauscat_save(n) / tot_tau
            if ( omega .ge. one ) omega = 0.99999d0
            VLIDORT_Optical_inputs%TS_deltau_vert_input(n) = tot_tau
            VLIDORT_Optical_inputs%TS_omega_total_input(n) = omega
          enddo

!  Call VLIDORT master (no linearizations)

          CALL VLIDORT_master ( &
          VLIDORT_FBoolean_inputs, &
          VLIDORT_MBoolean_inputs, &
          VLIDORT_FControl_inputs, &
          VLIDORT_MControl_inputs, &
          VLIDORT_Beam_inputs,     &
          VLIDORT_User_inputs,     &
          VLIDORT_Chapman_inputs,  &
          VLIDORT_Optical_inputs,  &
          VLIDORT_Write_inputs,    &
          VLIDORT_BRDF_inputs,     &
          VLIDORT_Outputs,         &
          VLIDORT_Status )

!  Exception handling
!  Exception handling

          CALL VLIDORT_WRITE_STATUS ( &
            'NO2_Execution.log', 35, OPENFILEFLAG, &
            VLIDORT_Status%TS_STATUS_INPUTCHECK,   &
            VLIDORT_Status%TS_NCHECKMESSAGES,      &
            VLIDORT_Status%TS_CHECKMESSAGES,       &
            VLIDORT_Status%TS_ACTIONS,             &
            VLIDORT_Status%TS_STATUS_CALCULATION,  &
            VLIDORT_Status%TS_MESSAGE,             &
            VLIDORT_Status%TS_TRACE_1,             &
            VLIDORT_Status%TS_TRACE_2,             &
            VLIDORT_Status%TS_TRACE_3 )

!  Stop if calculation error

          IF ( VLIDORT_Status%TS_STATUS_CALCULATION.ne.vlidort_success ) then
            write(*,*)'Finite-difference Run aborted for layer number ',k,'; cycle number ',ip,'; Look at error Log file'
            Stop
          ENDIF

!  write up FD-check results

          WRITE(57,'(1p12e14.5)') &
             half*(HEIGHT_GRID(k)+HEIGHT_GRID(k-1)),pres_save(k), gas_tau_data(k),isave2, &
             (VLIDORT_Outputs%TS_STOKES(1,1,1,UPIDX)-isave2)/EPSILON,                  &
              VLIDORT_LPOutputs%TS_PROFILEWF(1,K,1,1,1,UPIDX),&
             (VLIDORT_Outputs%TS_STOKES(1,1,1,UPIDX)-isave2)/EPSILON/gas_tau_data(k)/isave2,&
              VLIDORT_LPOutputs%TS_PROFILEWF(1,K,1,1,1,UPIDX)/isave2/gas_tau_data(k)

!  End FD layer check

        enddo

!  End pressure cycle

      enddo

!  4. Final section
!  ================
!  Close output file and move

      close(57)
      call system('mv fort.57 wfcld46_tc05_a05_no2_new')

!  Finish

      stop
END PROGRAM NO2AMF_AIRCRAFT_MAIN
